#include <system.h>
#include "ekboostload.h"
/**********************************************************
* EKBoostLoader 
*
*	BoostC PIC16F87X Bootloader utility.
*
*	Written by ryan@electronikits.com 
*
*	Revision 0.0.0.4
*	History:
*		Nov 30 2004 - Created
*		Dec 2  2004 - Bootloader now exits with nothing on stack.
*					  Fixed Addressing Bug
*					  Software Flow Control ( XON/XOFF )
*					  Added Support For other partnumbers
*					  Added Comments.
*		Dec 3  2004 - Added Configuration Byte Initialization
*		Dec 8  2004 - Optimized To Fit into 1 page of program memory
*   NOTES:
*
**********************************************************/

#define COMM_TIMEOUT 50000	//2 seconds

#pragma CLOCK_FREQ CLOCK_SPEED
#pragma DATA 0x2007, _HS_OSC & _WDT_OFF & _PWRTE_ON & _LVP_OFF & _BODEN_ON & _WRT_ENABLE_ON

//Linker Must be passed the following parameters
// to start bootloader at 0x1E05
// to start bootloader at 0x1F05 (7941) for 8K PIC 
// to start bootloader at 0x0F05 (3845) for 4K PIC
// to start bootloader at 0x0705 (1797) for 2K PIC
// -rb 7941

//Boot Jump Vector (To Bootloader..) @ 0x1E00
//This is the first executed code block and it jumps straight to the bootloader.
#if defined( _16F877 ) || defined( _16F876 )
	#pragma DATA 0x0000, 0x301F, 0x008A, 0x3005, 0x0082
#elif defined( _16F874 ) || defined( _16F873 )
	#pragma DATA 0x0000, 0x300F, 0x008A, 0x3005, 0x0082
#elif defined( _16F871 ) || defined( _16F870 )
	#pragma DATA 0x0000, 0x3007, 0x008A, 0x3005, 0x0082
#endif
//	movlw	0x1F  or 0x0F, or 0x07
//	movwf	pclath
//	movlw	0x05
//  movwf	pcl
//////////////////////////////////////////////////////////////////////////////////

//This is where the first 4 instructions of the application code are relocated to.
//#pragma DATA BOOTLOAD_JUMP, 0x018A, 0x0000, 0x018A, 0x0082
#if defined( _16F877 ) || defined( _16F876 )
	#pragma DATA 0x1F00, 0x018A, 0x0000, 0x0000, 0x018A, 0x0082
#elif defined( _16F874 ) || defined( _16F873 )
	#pragma DATA 0x0F00, 0x018A, 0x0000, 0x0000, 0x018A, 0x0082
#elif defined( _16F871 ) || defined( _16F870 )
	#pragma DATA 0x0700, 0x018A, 0x0000, 0x0000, 0x018A, 0x0082
#endif
//  clrf    pclath
//	nop				//THESE LAST 4 INSTRUCTIONS ARE OVERWRITTEN BY APPLICATION.	
//  nop		
//	clrf    pclath	
//	clrf    pcl
//////////////////////////////////////////////////////////////////////////////////

unsigned char hiAddr;
unsigned char loAddr;
unsigned char hiData;
unsigned char loData;
unsigned char bLoading;
unsigned char idx;
unsigned char length;
unsigned char rcType;
unsigned char a;
unsigned char b;
unsigned char c;
	
void main( )
{	
	//Initialize Serial Port
	g_CommTimeOut = 0;
	txsta = 0x24;		//Enable Tx, BRGH=1
	rcsta = 0x90;		//Enable SPort, Enable Rx
	spbrg = BR_CONST;	//9600 Baud

	//Send Prompt For Start.
	c = '>';
	serial_putchar( );

	//Receive Hex Records..
	bLoading = 1;
	while( bLoading )
	{
		serial_getbyte();
		if( g_CommTimeOut != 0x00 )
		{	//Exit Loop and continue to Application code jump.
			bLoading = 0;
		}
		else if (b == ':')
		{	// Example Hex Record
			//  :1000100008003C3023020318112830302302A300CB
			// 	:10 0010 00 08 00 3C 30 23 02 03 18 11 28 30 30 23 02 A3 00 CB 
			//	 ln addr cd d1  2  3  4  5  6  7  8  9 10 11 12 13 14 15 16 cs
			
			//Get Record Length
			serial_getbyte();
			a = b;
			serial_getbyte();
			length = ascii2byte( );
			//Get Address
			serial_getbyte();
			a = b;
			serial_getbyte();
			hiAddr = ascii2byte( );
			serial_getbyte();
			a = b;
			serial_getbyte();
			loAddr = ascii2byte( );
			//Get Record Type
			serial_getbyte();
			a = b;
			serial_getbyte();
			rcType = ascii2byte( );			
			
			if( rcType == 0x01 )	//If Last Record..
			{
				bLoading = 0;
			}
			else if( rcType == 0x00 )	//If Data Record, write to flash.
			{	//Dont write over loader.
				if (hiAddr < ( ( BOOTLOAD_START >> 8 ) & 0x00FF ) ) 
				{				
					//Shift Address Once To Right..
					//Because we are addressing every 2 bytes, not every byte.	
					loAddr >>= 1;
					if( hiAddr & 0x01 )
						loAddr |= 0x80;
					hiAddr = hiAddr >> 1;				

					//Retrieve All Data and Write To Flash
					idx = 0;
					while (idx < length)
					{
						serial_getbyte();
						a = b;
						serial_getbyte();
						loData = ascii2byte( );						
						serial_getbyte();
						a = b;
						serial_getbyte();	
						c = XOFF;					
						serial_putchar( );			//Pause Transmission
						hiData = ascii2byte( );

						if( hiAddr == 0 && loAddr < 4 ) 
						{	//Relocate Boot Sector Data To Upper Flash
							eeadr = loAddr+BL_JUMP_LOBYTE;
							eeadrh = BL_JUMP_HIBYTE;	
						}
						else
						{
							eeadr = loAddr;
							eeadrh = hiAddr;
						}	

						eedata = loData;
						eedath = hiData;
						write_flash( );

						//Jump to next 2 bytes.
						idx+=2;
						
						//Increment Address.
						loAddr++;
						if (loAddr == 0) hiAddr++;		
						c = XON;				
						serial_putchar( );			//Resume Transmission
					}
				}			
			}
		}
	}

	//Too Late.. Jumping To App. Or Restarting Bootloader.
	asm
	{
		//Set pclath
#if defined( _16F877 ) || defined( _16F876 )
		movlw	0x1F
#elif defined( _16F874 ) || defined( _16F873 )
		movlw	0x0F
#elif defined( _16F871 ) || defined( _16F870 )
		movlw	0x07
#endif	
		movwf	_pclath		
		//set pcl
		clrf	_pcl
	}	

	//Should Not Get Here.
}

//void write_flash( unsigned char hiAddr, unsigned char loAddr, unsigned char hiData, unsigned char loData )
void write_flash( void )
{	
	set_bit( eecon1, EEPGD );
	set_bit( eecon1, WREN );
	clear_bit( intcon, GIE );
	
	eecon2 = 0x55;
	eecon2 = 0xAA;
	set_bit( eecon1, WR );
	nop();
	nop();
	nop();
	clear_bit( eecon1, WR );
}

//Converts Two Ascii characters to its hex equivalent
unsigned char ascii2byte( void )
{
	if ( a < 0x3C )
		a = a - 0x30;
	else
		a = a - 0x37;
		
	if ( b < 0x3C )
		b = b - 0x30;
	else
		b = b - 0x37;
		
	return ( ( a << 4 ) + b );	
}

//Writes Character to Serial Port
void serial_putchar( )
{
	while( ( txsta & 0x02 ) == 0 );
	txreg = c;
}

//Receives Character from Serial Port
//Check g_CommTimeOut global to check for timeout.
void serial_getbyte( )
{
	unsigned int nTimeOut;	
	nTimeOut = 0;	
	while( ( pir1 & 0x20 ) == 0 )
	{
		delay_us( 40 );		
		nTimeOut++;
		if( nTimeOut > COMM_TIMEOUT )
		{	//Comm TimeOut Occurred
			g_CommTimeOut = 1;	
			return;
		}
	}		
	b = rcreg;
}

